package polyglot.ext.jl5.types;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import polyglot.ext.jl.types.ParsedClassType_c;
import polyglot.frontend.Source;
import polyglot.types.LazyClassInitializer;
import polyglot.types.MethodInstance;
import polyglot.types.Named;
import polyglot.types.PrimitiveType;
import polyglot.types.Resolver;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.types.TypeObject;
import polyglot.types.TypeSystem;
import polyglot.util.InternalCompilerError;
import polyglot.util.TypedList;

public class JL5ParsedClassType_c extends ParsedClassType_c implements JL5ParsedClassType, SignatureType{
    
    protected List enumConstants;
    // these are annotation elements in the annotation type
    protected List annotationElems;
  
    // these are annotations that have been declared on (applied to) the type
    protected List annotations;
    protected List<TypeVariable> typeVariables;
//    protected List typeArguments;

    public JL5ParsedClassType_c( TypeSystem ts, LazyClassInitializer init, Source fromSource){
        super(ts, init, fromSource);
    }
    
    public void annotations(List annotations){
        this.annotations = annotations;
    }

    public List annotations(){
        return annotations;
    }
        
    
    public void addEnumConstant(EnumInstance ei){
        enumConstants().add(ei);
    }

    public void addAnnotationElem(AnnotationElemInstance ai){
        annotationElems().add(ai);
    }

    public List enumConstants(){
        if (enumConstants == null){
            enumConstants = new TypedList(new LinkedList(), EnumInstance.class, false);
            ((JL5LazyClassInitializer)init).initEnumConstants(this);
            freeInit();
        }
        return enumConstants;
    }
   
    protected void freeInit(){
        
    }
    
    
    public List annotationElems(){
        if (annotationElems == null){
            annotationElems = new TypedList(new LinkedList(), AnnotationElemInstance.class, false);
            ((JL5LazyClassInitializer)init).initAnnotations(this);
            freeInit();
        }
        return annotationElems;
    }

    protected boolean initialized(){
        return super.initialized() && this.enumConstants != null && this.annotationElems != null;
    }
    
    public EnumInstance enumConstantNamed(String name){
        for(Iterator it = enumConstants().iterator(); it.hasNext();){
            EnumInstance ei = (EnumInstance)it.next();
            if (ei.name().equals(name)){
                return ei;
            }
        }
        return null;
    }
    
    public AnnotationElemInstance annotationElemNamed(String name){
        for(Iterator it = annotationElems().iterator(); it.hasNext();){
            AnnotationElemInstance ai = (AnnotationElemInstance)it.next();
            if (ai.name().equals(name)){
                return ai;
            }
        }
        return null;
    }

    public void addMethod(MethodInstance mi){
        if (JL5Flags.isAnnotationModifier(flags())){
            //addAnnotationElem(ts.annotationElemInstance(mi.position(), this, mi.flags(), mi.type(), mi,name(), 
                        
        }
        super.addMethod(mi);
    }

	// find methods with compatible name and formals as the given one
    public List methods(JL5MethodInstance mi) {
        List l = new LinkedList();

        for (Object o : methodsNamed(mi.name())) {
            JL5ProcedureInstance pi = (JL5ProcedureInstance) o;
            if (pi.hasSameFormals(mi)) {
                l.add(pi);
            }
        }
	return l;
    }
    
    public boolean isImplicitCastValidImpl(Type toType){
        if (isAutoUnboxingValid(toType)) return true;
        return super.isImplicitCastValidImpl(toType);
    }

    private boolean isClassToIntersectionValid(Type toType){
        TypeVariable it = (TypeVariable)toType;
        if (it.bounds() == null || it.bounds().isEmpty()) return true;
        return ts.isImplicitCastValid(this, it.bounds().get(0));
    }
    
    private boolean isAutoUnboxingValid(Type toType){
        if (!toType.isPrimitive()) return false;
        if (this.fullName().equals("java.lang.Integer")) return ts.isImplicitCastValid(ts.Int(), toType);
        if (this.fullName().equals("java.lang.Boolean")) return ts.isImplicitCastValid(ts.Boolean(), toType);
        if (this.fullName().equals("java.lang.Byte")) return ts.isImplicitCastValid(ts.Byte(), toType);
        if (this.fullName().equals("java.lang.Short")) return ts.isImplicitCastValid(ts.Short(), toType);
        if (this.fullName().equals("java.lang.Character")) return ts.isImplicitCastValid(ts.Char(), toType);
        if (this.fullName().equals("java.lang.Long")) return ts.isImplicitCastValid(ts.Long(), toType);
        if (this.fullName().equals("java.lang.Float")) return ts.isImplicitCastValid(ts.Float(), toType);
        if (this.fullName().equals("java.lang.Double")) return ts.isImplicitCastValid(ts.Double(), toType);
        /*if (toType.isInt() && this.fullName().equals("java.lang.Integer")) return true;
        if (toType.isBoolean() && this.fullName().equals("java.lang.Boolean")) return true;
        if (toType.isByte() && this.fullName().equals("java.lang.Byte")) return true;
        if (toType.isShort() && this.fullName().equals("java.lang.Short")) return true;
        if (toType.isChar() && this.fullName().equals("java.lang.Character")) return true;
        if (toType.isLong() && this.fullName().equals("java.lang.Long")) return true;
        if (toType.isDouble() && this.fullName().equals("java.lang.Double")) return true;
        if (toType.isFloat() && this.fullName().equals("java.lang.Float")) return true;*/
        return false;
    }

    public List<TypeVariable> typeVariables(){
        return typeVariables;
    }
    
    // this is used when coming from source files
    public void addTypeVariable(TypeVariable type){
        if (typeVariables == null){
            typeVariables = new ArrayList<TypeVariable>();
        }
        if (!typeVariables.contains(type)){
            typeVariables.add(type);
            type.declaringClass(this);
        }
    }

    // this is used when coming from class files
    public void typeVariables(List<TypeVariable> vars){
        typeVariables = vars;
        for (TypeVariable variable : vars) {
            variable.declaringClass(this);
        }
    }
    
    public boolean hasTypeVariable(String name){
        if (typeVariables == null || typeVariables.isEmpty()) return false;
        for (Iterator<TypeVariable> it = typeVariables.iterator(); it.hasNext(); ){
            TypeVariable iType = it.next();
            if (iType.name().equals(name)) return true;
        }
        return false;
    }

    public TypeVariable getTypeVariable(String name){
        for (Iterator<TypeVariable> it = typeVariables.iterator(); it.hasNext(); ){
            TypeVariable iType = it.next();
            if (iType.name().equals(name)) return iType;
        }
        return null;
    }

    public boolean isGeneric(){
        if ((typeVariables != null) && !typeVariables.isEmpty()) return true;
        return false;
    }

    // this is only for debugging or something
    public String toString(){
        StringBuffer sb = new StringBuffer(super.toString());
        if ((typeVariables != null) && !typeVariables.isEmpty()){
            sb.append("<");
            sb.append(typeVariables);
            sb.append(">");
        }
        return sb.toString();
    }

    public LazyClassInitializer init(){
        return this.init;
    }

    /*public List typeArguments(){
        return typeArguments;
    }

    public void typeArguments(List args){
        this.typeArguments = args;
    }
   
    public boolean isParameterized(){
        if (typeArguments != null && !typeArguments.isEmpty()) return true;
        return false;
    }
    
    public String translate(Resolver c){
        StringBuffer sb = new StringBuffer(super.translate(c));
        if (isParameterized()){
            sb.append("<");
            for (Iterator it = typeArguments().iterator(); it.hasNext(); ){
                sb.append(((Type)it.next()).translate(c));
                if (it.hasNext()){
                    sb.append(", ");
                }
            }
            sb.append(">");
        }
        return sb.toString();
    }*/

    public boolean numericConversionValidImpl(long l){
        return numericConversionValid(new Long(l));
    }

    public boolean numericConversionValidImpl(Object value){
        if (value == null) return false;

        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (value instanceof Float){
            //if (ts.equals(this, ts.FloatWrapper())) return true;
            return false;
        }
        if (value instanceof Double){
            //if (ts.equals(this, ts.DoubleWrapper())) return true;
            return false;
        }

        long v;
        if (value instanceof Number){
            v = ((Number) value).longValue();
        }
        else if (value instanceof Character){
            v = ((Character) value).charValue();
        }
        else {
            return false;
        }

        if (ts.equals(this, ts.LongWrapper()) && value instanceof Long) return true;
        if (ts.equals(this, ts.IntegerWrapper()) && value instanceof Integer) return Integer.MIN_VALUE <= v && v <= Integer.MAX_VALUE;
        if (ts.equals(this, ts.CharacterWrapper()) && value instanceof Character) return Character.MIN_VALUE <= v && v <= Character.MAX_VALUE;
        if (ts.equals(this, ts.ShortWrapper()) && value instanceof Short) return Short.MIN_VALUE <= v && v <= Short.MAX_VALUE;
        if (ts.equals(this, ts.ByteWrapper()) && value instanceof Byte) return Byte.MIN_VALUE <= v && v <= Byte.MAX_VALUE;

        return false;
    }


// 	// a straw man?  two class types are equal if they have the same name
//     public boolean equalsImpl(TypeObject t) {
// 	System.out.println("comparing " + this + " and " + t);
// 	if (t instanceof JL5ParsedClassType) {
// 	    JL5ParsedClassType ct = (JL5ParsedClassType) t;
// 	    return this.fullName().equals(ct.fullName());
// 	}
// 	return false;
//     }
    
    /*public boolean equalsImpl(TypeObject t){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (t instanceof PrimitiveType){ 
            if (this.isBoolean() && ((Type)t).isBoolean()) return true;
            if (this.isInt() && ((Type)t).isInt()) return true;
            if (this.isByte() && ((Type)t).isByte()) return true;
            if (this.isShort() && ((Type)t).isShort()) return true;
            if (this.isChar() && ((Type)t).isChar()) return true;
            if (this.isLong() && ((Type)t).isLong()) return true;
            if (this.isFloat() && ((Type)t).isFloat()) return true;
            if (this.isDouble() && ((Type)t).isDouble()) return true;
        }
        return super.equalsImpl(t);
    }*/

    public boolean equivalentImpl(TypeObject t){
   //     JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (t instanceof PrimitiveType){ 
            if (this.isBoolean() && ((Type)t).isBoolean()) return true;
            if (this.isInt() && ((Type)t).isInt()) return true;
            if (this.isByte() && ((Type)t).isByte()) return true;
            if (this.isShort() && ((Type)t).isShort()) return true;
            if (this.isChar() && ((Type)t).isChar()) return true;
            if (this.isLong() && ((Type)t).isLong()) return true;
            if (this.isFloat() && ((Type)t).isFloat()) return true;
            if (this.isDouble() && ((Type)t).isDouble()) return true;
        }
        return false;
    }

    public boolean isNumeric(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.IntegerWrapper())) return true;
        if (this.equals(ts.ByteWrapper())) return true;
        if (this.equals(ts.ShortWrapper())) return true;
        if (this.equals(ts.CharacterWrapper())) return true;
        if (this.equals(ts.LongWrapper())) return true;
        if (this.equals(ts.FloatWrapper())) return true;
        if (this.equals(ts.DoubleWrapper())) return true;
        return false;
    }

    public boolean isByte(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.ByteWrapper())) return true;
        return false;
    }

    public boolean isShort(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.ShortWrapper())) return true;
        return false; 
    }

    public boolean isIntOrLess(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.IntegerWrapper())) return true;
        if (this.equals(ts.ByteWrapper())) return true;
        if (this.equals(ts.ShortWrapper())) return true;
        if (this.equals(ts.CharacterWrapper())) return true;
        return false; 
    }

    public boolean isInt(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.IntegerWrapper())) return true;
        return false;
    }

    public PrimitiveType toPrimitive() {
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.IntegerWrapper())) return typeSystem().Int();
        if (this.equals(ts.ByteWrapper())) return typeSystem().Byte();
        if (this.equals(ts.ShortWrapper())) return typeSystem().Short();
        if (this.equals(ts.CharacterWrapper())) return typeSystem().Char();
        if (this.equals(ts.LongWrapper())) return typeSystem().Long();
        if (this.equals(ts.FloatWrapper())) return typeSystem().Float();
        if (this.equals(ts.DoubleWrapper())) return typeSystem().Double();
        return super.toPrimitive();
    }

    public boolean isBoolean(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.BooleanWrapper())) return true;
        return false;
    }

    public boolean isChar(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.CharacterWrapper())) return true;
        return false;
    }

    public boolean isFloat(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.FloatWrapper())) return true;
        return false;
    }

    public boolean isLong(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.LongWrapper())) return true;
        return false;
    }
    
    public boolean isDouble(){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (this.equals(ts.DoubleWrapper())) return true;
        return false;
    }


    public boolean isCastValidImpl(Type toType){
        JL5TypeSystem ts = (JL5TypeSystem)typeSystem();
        if (isNumeric() || isBoolean()){
            if (isBoolean() && toType.isBoolean()) return true;
            if (isByte() && (toType.isByte() || toType.isShort() || toType.isInt() || toType.isLong() || toType.isFloat() || toType.isDouble())) return true;
            if (isShort() && (toType.isShort() || toType.isInt() || toType.isLong() || toType.isFloat() || toType.isDouble())) return true;
            if (isChar() && (toType.isChar() || toType.isInt() || toType.isLong() || toType.isFloat() || toType.isDouble())) return true;
            if (isInt() && (toType.isInt() || toType.isLong() || toType.isFloat() || toType.isDouble())) return true;
            if (isLong() && (toType.isLong() || toType.isFloat() || toType.isDouble())) return true;
            if (isFloat() && (toType.isFloat() || toType.isDouble())) return true;
            if (isDouble() && toType.isDouble()) return true;

            //if (toType.isPrimitive()) return ts.isCastValid(toPrimitive(), toType);
        }
        return super.isCastValidImpl(toType);
    }
    
    public String translate(Resolver c) {
        if (isTopLevel()) {
            if (package_() == null) {
                return name();
            }

            // Use the short name if it is unique.
            if (c != null) {
                try {
                    Named x = c.find(name());

                    if (ts.equals(this, x)) {
                        return name();
                    }
                }
                catch (SemanticException e) {
                }
            }

            return package_().translate(c) + "." + name();
        }
        else if (isMember()) {
            // Use only the short name if the outer class is anonymous.
            if (container().toClass().isAnonymous()) {
                return name();
            }

            // Use the short name if it is unique.
            if (c != null) {
                try {
                    Named x = c.find(name());

                    if (ts.equals(this, x) && !(container() instanceof ParameterizedType)) {
                        return name();
                    }
                }
                catch (SemanticException e) {
                }
            }

            return container().translate(c) + "." + name();
        }
        else if (isLocal()) {
            return name();
        }
        else {
            throw new InternalCompilerError("Cannot translate an anonymous class.");
        }
    }

    public String signature(){
        return "L"+fullName().replaceAll("\\.", "/")+";";
    }
}
